/****************************************************************
**
** Implementation of Gantt class
**
** This software is released under the GPL General Public License,
** see COPYING for details.
**
****************************************************************/

#include "gantt.h"

#include <qpainter.h>
#include <qprinter.h>
#include <qprintdialog.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qtextstream.h>
#include <qstring.h>
#include <qdatetime.h>
#include <qevent.h>


Gantt::Gantt( char *filenamearg, QWidget *parent, const char *name )
        : QWidget( parent, name )
	{
	setBackgroundColor( "white" );
	textpixx = 150;
	daypix = 3;
	fileisread = 0;
	isprinting = 0;
	file = new QFile();
	printer = new QPrinter();
	file->setName( filenamearg );
	if( file->exists() )
		{
		open( file );
		repaint();
		}
	}


void Gantt::open()
	{
	QString filename( QFileDialog::getOpenFileName( 0, "*.gantt" ) );
	if ( !filename.isNull() )             // if cancel was not pressed
		{
		file->setName( (const char *)filename );
		open( file );
		}
	}


void Gantt::print()  // draw gantt to printer
	{
	if( fileisread )    // if file reading went OK
		{
		if ( printer->setup() )   // if pushed OK button in printer setup
			{
			isprinting = 1;
			QPainter painter;
			painter.begin( printer );
			paint( &painter );      // calls the paint function
			painter.end();
			isprinting = 0;
			}
		}
	}


void Gantt::draw()  // draw gantt to screen
	{
	if( fileisread )   // if file reading went OK
		{
		QPainter painter;
		painter.begin( this );
		paint( &painter );
		painter.end();
		}
	}


void Gantt::refresh()           // when F5 is pushed
	{
	if( file->exists() )
		{
		open( file );
		repaint();
		}
	}


void Gantt::changetextpix( int tpix )
	{
	textpixx = tpix;
	repaint();
	}


void Gantt::changedaypix( int dpix )
	{
	daypix = dpix;
	repaint();
	}


void Gantt::open( QFile *file )
	{
	if( file->open(IO_ReadOnly) )
		{
		QTextStream t( file );
		readfile( &t );
		fileisread = 1;
		}
	else
		{
		printf( "File not found\n" );
		fileisread = 0;
		}
	}


void Gantt::readfile( QTextStream *t )  // reads gantt file, used by print() and draw()
	{
	QString qstr1, qstr2, yyyy, mm, dd;
	numtasks=0;
	int pos1, pos2, pos3;
	
	while ( !t->eof() && numtasks<NUMTASKS )     // loops through the file
		{
		tasks[numtasks] = new task;
		qstr1 = t->readLine();
		
		if( !qstr1.isEmpty() && qstr1.left(1) != "#" )  // if not an empty line and not a comment
			{
			if( qstr1.left(1) == "%" )    // if a header line
				{
				pos1 = qstr1.find( '%', 1, TRUE );
				qstr2 = qstr1.mid( 1, pos1-1 );
				
				pos2 = qstr2.find( '-', 0, TRUE );
				dd = qstr2.left(pos2);
				pos3 = qstr2.find( '-', pos2+1, TRUE );
				mm = qstr2.mid( pos2+1, pos3-pos2-1 );
				yyyy = qstr2.right( qstr2.length() - pos3-1 );
				project_startdate = new QDate( yyyy.toInt(), mm.toInt(), dd.toInt() );
				
				qstr2 = qstr1.right( qstr1.length() - pos1 - 1 );
				header = new QString( qstr2 );
				}
			else                   // if a task line
				{
				// num
				pos1 = qstr1.find( ':', 0, TRUE );
				qstr2 = qstr1.left(pos1);
				tasks[numtasks]->num = qstr2.toInt();
				
				// type
				pos2 = qstr1.find( ':', pos1+1, TRUE );
				qstr2 = qstr1.mid( pos1+1, pos2-pos1-1 );
				tasks[numtasks]->type = new QString( qstr2 );
				
				// startdate
				pos1 = qstr1.find( ':', pos2+1, TRUE );
				qstr2 = qstr1.mid( pos2+1, pos1-pos2-1 );
				pos3 = qstr2.find( '-', 0, TRUE );
				dd = qstr2.left(pos3);
				pos2 = qstr2.find( '-', pos3+1, TRUE );
				mm = qstr2.mid( pos3+1, pos2-pos3-1 );
				yyyy = qstr2.right( qstr2.length() - pos2-1 );
				tasks[numtasks]->startdate = new QDate( yyyy.toInt(), mm.toInt(), dd.toInt() );
				
				// enddate
				pos2 = qstr1.find( ':', pos1+1, TRUE );
				if( tasks[numtasks]->type->copy() != "mile" )  // if not a milestone
					{
					qstr2 = qstr1.mid( pos1+1, pos2-pos1-1 );
					pos3 = qstr2.find( '-', 0, TRUE );
					dd = qstr2.left(pos3);
					pos1 = qstr2.find( '-', pos3+1, TRUE );
					mm = qstr2.mid( pos3+1, pos1-pos3-1 );
					yyyy = qstr2.right( qstr2.length() - pos1-1 );
					tasks[numtasks]->enddate = new QDate( yyyy.toInt(), mm.toInt(), dd.toInt() );
					}
				else         // milestones don't need enddate, so a blank entry is ok
					tasks[numtasks]->enddate = new QDate( tasks[numtasks]->startdate->addDays(0) );
				
				// after
				pos1 = qstr1.find( ':', pos2+1, TRUE );
				qstr2 = qstr1.mid( pos2+1, pos1-pos2-1 );
				tasks[numtasks]->after = qstr2.toInt();
				
				//during
				pos2 = qstr1.find( ':', pos1+1, TRUE );
				qstr2 = qstr1.mid( pos1+1, pos2-pos1-1 );
				tasks[numtasks]->during = qstr2.toInt();
				
				//description
				qstr2 = qstr1.right( qstr1.length() - pos2-1 );
				tasks[numtasks]->description = new QString( qstr2 );
				
				numtasks++;
				}
			}
		}
	file->close();
	}


void Gantt::paint( QPainter *painter )  // draw gantt, used by print() and draw()
	{
	int page=1, pagewidth, pageheight;
	if( isprinting )
		{
		pagewidth = 600;
		pageheight = 400;
		}
	else
		{
		pagewidth = width();
		pageheight = height();
		}
	
	
	// Draws the header text
	painter->setPen( "black" );
	painter->setFont( QFont("times", 20, QFont::Bold) );
	painter->drawText( 10, 0, pagewidth-20, TEXTPIXY-20, AlignCenter, header->copy() );
	
	
	int i=0, pagetasks=20, currentpage;
	while( i<numtasks )
		{
		currentpage = page;
		painter->setPen( "darkgrey" );
		painter->drawLine( 0, TEXTPIXY-20, pagewidth, TEXTPIXY-20 );
		painter->drawLine( 0, TEXTPIXY, pagewidth, TEXTPIXY );
		// Draws lines for weeks, months and current date
		painter->setFont( QFont("helvetica", FONTSIZE, QFont::Normal) );
		QDate tempdate = project_startdate->addDays(0);
		for( int day=textpixx; day < pagewidth; day = day+daypix )
			{
			if( tempdate.dayOfWeek() == 1 )  // if a monday
				{
				QPen dottedpen( "lightgrey", 0, DashLine );
				painter->setPen ( dottedpen );
				painter->drawLine( day, TEXTPIXY, day, pageheight );  // Draws a dotted line for each week
				}
			if( tempdate.day() == 1 )    // if first day of month
				{
				painter->setPen( "darkgrey" );
				painter->drawLine( day, TEXTPIXY-20, day, pageheight );  // Draws a solid line for each month
				painter->setPen( "black" );
				switch( tempdate.month() )
					{
					case 1: painter->drawText( day+5, TEXTPIXY-5, "January"); break;
					case 2: painter->drawText( day+5, TEXTPIXY-5, "February"); break;
					case 3: painter->drawText( day+5, TEXTPIXY-5, "March"); break;
					case 4: painter->drawText( day+5, TEXTPIXY-5, "April"); break;
					case 5: painter->drawText( day+5, TEXTPIXY-5, "May"); break;
					case 6: painter->drawText( day+5, TEXTPIXY-5, "June"); break;
					case 7: painter->drawText( day+5, TEXTPIXY-5, "July"); break;
					case 8: painter->drawText( day+5, TEXTPIXY-5, "August"); break;
					case 9: painter->drawText( day+5, TEXTPIXY-5, "September"); break;
					case 10: painter->drawText( day+5, TEXTPIXY-5, "October"); break;
					case 11: painter->drawText( day+5, TEXTPIXY-5, "November"); break;
					case 12: painter->drawText( day+5, TEXTPIXY-5, "December"); break;
					}
				}
			if( tempdate == tempdate.currentDate() )    // if current date
				{
				painter->setPen( "red" );
				painter->drawLine( day, TEXTPIXY, day, pageheight );  // Draws a solid red line for the current date
				}
			tempdate = tempdate.addDays(1);
			}
		
		// Draws the tasks, mains and milestones
		for( i; i<numtasks && page==currentpage; i++ )  // for each task
			{
			if( tasks[i] )   // if this tasknumber is read from file
				{
				int j=0;
				if( tasks[i]->after != 0 )   // if 'after' is specified (which means the task depends on other tasks)
					{
					while( tasks[j]->num != tasks[i]->after && j < i )
						{ j++; }
					}
				
				int jbegin, jend;
				if( (int)( (float)(i/pagetasks)-(float)(j/pagetasks) ) > 0 )
					{
					jbegin = 0;
					jend = 1200;
					}
				else
					{
					jbegin = j%pagetasks;
					jend = j%pagetasks;
					}
				
				if( tasks[i]->type->copy() == "task" )   // draws a normal task
					{
					painttask( painter, i%pagetasks );
					if( tasks[i]->after != 0 )                     // if 'after' is specified (which means the task depends on other tasks)
						{ painttaskendarrow( painter, i%pagetasks, jend ); }   // draws first part of the dependency arrow
					}
				else if( tasks[i]->type->copy() == "main" )   // draws a main task
					{
					paintmain( painter, i%pagetasks );
					if( tasks[i]->after != 0 )                     // if 'after' is specified (which means the task depends on other tasks)
						{ paintmainendarrow( painter, i%pagetasks, jend ); }   // draws first part of the dependency arrow
					}
				else if( tasks[i]->type->copy() == "mile" )  // draws a milestone
					{
					paintmile( painter, i%pagetasks );
					if( tasks[i]->after != 0 )                     // if 'after' is specified (which means the task depends on other tasks)
						{ paintmileendarrow( painter, i%pagetasks, jend ); }   // draws first part of the dependency arrow
					}
				
				if( tasks[j]->type->copy() == "task" && tasks[i]->after != 0 )   // draws a normal task
					{
					painttaskbeginarrow( painter, i%pagetasks, jbegin );   // draws first part of the dependency arrow
					}
				else if( tasks[j]->type->copy() == "main" && tasks[i]->after != 0 )   // draws a main task
					{
					paintmainbeginarrow( painter, i%pagetasks, jbegin );   // draws first part of the dependency arrow
					}
				else if( tasks[j]->type->copy() == "mile" && tasks[i]->after != 0 )  // draws a milestone
					{
					paintmilebeginarrow( painter, i%pagetasks, jbegin );   // draws first part of the dependency arrow
					}
				}
			if( !i%pagetasks && isprinting )
				{
				printer->newPage();
				page++;
				}
			}
		}
	}


void Gantt::paintEvent( QPaintEvent *e )  // repaints automagically when needed
	{
	draw();
	}


void Gantt::painttask( QPainter *painter, int i )
	{
	painter->setPen( "black" );
	painter->setBrush( "lightblue" );
	painter->setFont( QFont("helvetica", FONTSIZE, QFont::Normal) );
	painter->drawText( 10, TEXTPIXY+13+i*SPACING,
	                   textpixx-10, 20, AlignLeft,
	                   tasks[i]->description->copy() );
	painter->drawRect( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	                   TEXTPIXY+10+i*SPACING,
	                   daypix * tasks[i]->startdate->daysTo( tasks[i]->enddate->addDays(0) ),
	                   15 );
	}


void Gantt::paintmain( QPainter *painter, int i )
	{
	painter->setPen( "black" );
	painter->setBrush( "black" );
	painter->setFont( QFont("helvetica", FONTSIZE, QFont::Bold) );
	painter->drawText( 10, TEXTPIXY+13+i*SPACING,
	                   textpixx-10, 20, AlignLeft,
	                   tasks[i]->description->copy() );
	QPointArray a;
	a.setPoints( 8,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) - 5,
	             TEXTPIXY+15+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->enddate->addDays(0) ) + 5,
	             TEXTPIXY+15+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->enddate->addDays(0) ) + 5,
	             TEXTPIXY+20+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->enddate->addDays(0) ),
	             TEXTPIXY+25+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->enddate->addDays(0) ) - 5,
	             TEXTPIXY+20+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) + 5,
	             TEXTPIXY+20+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	             TEXTPIXY+25+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) - 5,
	             TEXTPIXY+20+i*SPACING );
	painter->drawPolygon( a );
	}


void Gantt::paintmile( QPainter *painter, int i )
	{
	painter->setPen( "black" );
	painter->setBrush( "black" );
	painter->setFont( QFont("helvetica", FONTSIZE, QFont::Normal, 1 ) );
	painter->drawText( 10, TEXTPIXY+13+i*SPACING,
	                   textpixx-10, 20, AlignLeft,
	                   tasks[i]->description->copy() );
	QPointArray a;
	a.setPoints( 4,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	             TEXTPIXY+15+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) + 5,
	             TEXTPIXY+20+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	             TEXTPIXY+25+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) - 5,
	             TEXTPIXY+20+i*SPACING );
	painter->drawPolygon( a );
	QString qstr;
	painter->setFont( QFont("helvetica", FONTSIZE, QFont::Bold) );
	painter->drawText( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) + 10,
	                   TEXTPIXY+25+i*SPACING,
	                   qstr.setNum( tasks[i]->startdate->day() ) + "." + qstr.setNum( tasks[i]->startdate->month() ) );
	}


void Gantt::painttaskbeginarrow( QPainter *painter, int i, int j )
	{
	if ( tasks[j]->enddate->daysTo( tasks[i]->startdate->addDays(0) ) * daypix < 7 )
		{
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+17+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+17+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+17+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+27+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+27+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING );
		}
	else
		{
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+17+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+17+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+17+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING );
		}
	}


void Gantt::painttaskendarrow( QPainter *painter, int i, int j )
	{
	painter->setPen( "black" );
	painter->setBrush( "black" );
	painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	                   TEXTPIXY+27+j*SPACING,
	                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	                   TEXTPIXY+3+i*SPACING );
	QPointArray a;
	a.setPoints( 3,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) - 5,
	             TEXTPIXY+4+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) + 5,
	             TEXTPIXY+4+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	             TEXTPIXY+9+i*SPACING );
	painter->drawPolygon( a );
	}


void Gantt::paintmainbeginarrow( QPainter *painter, int i, int j )
	{
	if ( tasks[j]->enddate->daysTo( tasks[i]->startdate->addDays(0) ) * daypix < 7 )
		{
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+26+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING );
		}
	else
		{
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+18+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+18+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+18+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING );
		}
	}


void Gantt::paintmainendarrow( QPainter *painter, int i, int j )
	{
	painter->setPen( "black" );
	painter->setBrush( "black" );
	painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	                   TEXTPIXY+27+j*SPACING,
	                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	                   TEXTPIXY+8+i*SPACING );
	QPointArray a;
	a.setPoints( 3,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) - 5,
	             TEXTPIXY+9+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) + 5,
	             TEXTPIXY+9+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	             TEXTPIXY+14+i*SPACING );
	painter->drawPolygon( a );
	}


void Gantt::paintmilebeginarrow( QPainter *painter, int i, int j )
	{
	if ( tasks[j]->enddate->daysTo( tasks[i]->startdate->addDays(0) ) * daypix < 7 )
		{
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+20+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+20+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+20+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+27+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ) + 7,
		                   TEXTPIXY+27+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING );
		}
	else
		{
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[j]->enddate->addDays(0) ),
		                   TEXTPIXY+20+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+20+j*SPACING );
		painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+20+j*SPACING,
		                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
		                   TEXTPIXY+27+j*SPACING );
		}
	}


void Gantt::paintmileendarrow( QPainter *painter, int i, int j )
	{
	painter->setPen( "black" );
	painter->setBrush( "black" );
	painter->drawLine( textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	                   TEXTPIXY+27+j*SPACING,
	                   textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	                   TEXTPIXY+7+i*SPACING );
	QPointArray a;
	a.setPoints( 3,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) - 5,
	             TEXTPIXY+8+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ) + 5,
	             TEXTPIXY+8+i*SPACING,
	             textpixx + daypix * project_startdate->daysTo( tasks[i]->startdate->addDays(0) ),
	             TEXTPIXY+13+i*SPACING );
	painter->drawPolygon( a );
	}
